/*

    - continuous spin model = O(3) model,  XY=O(2) model

    (c) 1996-2004 J. Schulenburg

    see README, ChangeLog and TODO for general informations

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "config.h"   /* you should edit it */

#define version_string "1.39 Nov2004"

#if defined(CONFIG_XY)
#define XY    1
#else
#define XY    0		/* XY: 1=XY-model 0=O(3)-model */
#endif

int vvv=1;			/* verbose mode */

/*	define N, ww_get(i,j)=bond between i and j	*/

#include "other.inc"

unsigned long start=1000ul;
unsigned long steps=1000000ul;	/* best is more than 1M */

/*     gnuc                 CC-challenge */
#if defined(__GNUC__) || defined(__CPP__)
#define Static_Inline static inline
#else
#define Static_Inline static
#endif

#define MaxExp          600  /* 700  ??? maximum of exponent */
#define NJ		8	/* same as DimJJ */
#define inline

/* --- public variables: */
double jj[NJ+1][3]={ { 0.0, 0.0, 0.0 },
                     { 1.0, 1.0, 1.0 },  /* Jxi=Jyi=Jzi=isotropic */
                     { .25, .25, .25 },
                     { 0.1, 0.1, 0.1 }},
       gerg=0.0,gkt,beta,e_border;   /* gerg=energy, gkt=global_kT */
/* --- UpperBorder = - LowerBorder = e_border, 
       (e_min > (-e_border)) => frustration   --- */ 
volatile unsigned long anz;	/* MC-loops, out via 10s-INT */
double s[N][3],sum_sz;		/* spins and sum_sz */
int    wt[N],wp[N];		/* spin-vectors 0..MAXW=0..2*pi */
double hx=0.0;			/* hx-field Ham+hx*Sx */
FILE *f1=NULL;			/* output-file */
int column=0;                   /* number of column in outputfile */

/* non-0-elements of ww n=anzahl,a-b=index_bounds_of_idx,1-2=sites */
int ww_n,ww_a[N],ww_b[N],ww_i[8*N],ww_j[8*N],ww_r[8*N];
		/* taballierte Winkelfunktionen sin(M_PI*x) */

#ifdef __MSDOS__
#define MAXW  1024
#else
#define MAXW  32768 	/* Vollkreiswinkel = 2*M_PI */
#endif
#if ((MAXW-1) ^ ((MAXW-1)>>1)) != (MAXW >> 1)
#error MAXW must be a power of 2
#endif
double  sinx[MAXW],cosx[MAXW],		/* sin(rand(0..2pi))		*/
        siny[MAXW],cosy[MAXW],		/* sin(acos(rand(-1..1)))	*/
        logx[MAXW];


/* ======= signal-transfer mit USR1=0bit USR2=1bit zwischen programmen! ==== */
/* =======       USR1:fkt(i);i=0; USR2:i++;  ==== */
#ifndef __MSDOS__
#include <signal.h>
#include <unistd.h>

volatile int sig1;	/* set by signal handler */

void outE(FILE *ff);	/* declaration */

void my_handler(int sig){ static int i=0;
    switch(sig){
      case SIGUSR1: sig1=sig;/* cmd_i; */
                    if(i==0){ outE(stderr);fprintf(stderr,"\n"); }
                    i=0;break;
      case SIGUSR2: sig1=sig;i++;break;
      case SIGALRM: sig1=sig;alarm(10);
                      if(vvv & 4){ outE(stderr);fprintf(stderr,"\r");} break;
      default:      sig1=sig;fprintf(stderr,"\nsignal(%d) received ",sig);}
    signal(sig,my_handler); /* neu-umlenken */}

int ini_handler(int vvv){ /* long pid=getpid(); */
   sig1=0;
   signal(SIGUSR1,my_handler);
   signal(SIGUSR2,my_handler);
   signal(SIGALRM,my_handler);
   if(vvv & 4)alarm(10);     /* alle 10s (geht nicht auf LAND+CHALLENGE+...!?) */
   return 0;}
#else
 int ini_handler(int vvv){ return vvv;}
#endif
/* ================================================================== */

int ifNaN(double r){ return( (fabs(r)<1e+14)?0:1 );}  /* NaN ? */

void _randomize(int seed){
  srand((unsigned)((seed)?seed:time(NULL)));return;}

#if (RAND_MAX & 32767)!=32767
#error RAND_MAX must be 2^n-1
#endif
Static_Inline double _randreal(){ 	/*  x=[0,1)  RAND_MAX must be 2^n */
  static double rrm=1.0/RAND_MAX; /* mult. ist schneller als div. */
  return ( (rand()&RAND_MAX)*rrm ); } /* sun: rand()>RAND_MAX */
					
/* === s[a]*s[b] === */
Static_Inline  double prd(int a,int b){
  return(s[a][0]*s[b][0]+s[a][1]*s[b][1]+s[a][2]*s[b][2]);}

/* === s[a] (s[b] x s[c]) === */
Static_Inline  double spat(int a,int b,int c){
  return(s[a][0]*(s[b][1]*s[c][2]-s[b][2]*s[c][1])
        +s[a][1]*(s[b][2]*s[c][0]-s[b][0]*s[c][2])
        +s[a][2]*(s[b][0]*s[c][1]-s[b][1]*s[c][0]));}

Static_Inline  		/* (wt,wp)_i => s(x,y,z)_i */
void setspin(double s[3],unsigned int wt,unsigned int wp){
  wt&=(MAXW-1);
  wp&=(MAXW-1);
#if XY
  s[0] = cosx[wp];
  s[1] = sinx[wp];
  s[2] = 0; 
#else
  s[0] = siny[wt] * cosx[wp];
  s[1] = siny[wt] * sinx[wp];
  s[2] = cosy[wt]; 
#endif  
}

Static_Inline  		/* i => (wt,wp)_i => s(x,y,z)_i */
void setrandomspin(int i){
  wt[i] = rand()&(MAXW-1);
  wp[i] = rand()&(MAXW-1);
  setspin(s[i],wt[i],wp[i]); }

		/* s(x,y,z) => rr*(wt,wp) */
void kart2kug(double *s,double *rr,int *wt,int *wp){
  double st,t,p,r;
  r=sqrt(s[0]*s[0]+s[1]*s[1]+s[2]*s[2]);
  if(r<1e-14){p=0.0;t=0.0;r=0.0;}
  else {
    t=acos(s[2]/r);		/* 0..M_PI */
    st=sin(t);t/=M_PI;
    if(fabs(st)<1e-15)p=0.0;
    else if( fabs(s[0]/st/r)>1.0 ){p=0.0;if(t<0.5)t=0.0; else t=1.0;}
    else { p=acos(s[0]/st/r)/M_PI;if(s[1]<0.0)p=(-p); }
    /* if(ifNaN(p))mybreak(); */
    while(t>=2.0)t-=2.0;while(t< 0.0)t+=2.0;if(t>1.0){ t=2.0-t;p+=1.0; }
    while(p>=2.0)p-=2.0;while(p< 0.0)p+=2.0;
  }
  if(wt)(*wt)=((unsigned int)(.5*t*MAXW))&(MAXW-1);
  if(wp)(*wp)=((unsigned int)(.5*p*MAXW))&(MAXW-1);;
  if(rr)(*rr)=r;  }

/* new version of norm_angles Mar99
 * input new axis in old koordinates (3x3 matrix)
 * ACHTUNG bei Anisotropie !???
 */
 void norm_angles(double hx){	/* wt ca. 0..1, wp 0..2 [Pi] */
    double ss[3],r,a,xyz[3][3]={{1,0,0},{0,1,0},{0,0,1}}; int i,j;
    /* --- set x axis equal to spin 0 if hx==0 */
    if( hx==0 )for(i=0;i<3;i++) xyz[0][i]=s[0][i];
    /* --- y axis is s1 orthogonalized to x=s0 */
    for(j=0;j<N;j++){
      for(a=0.0,i=0;i<3;i++) a+=xyz[0][i]*s[j][i];
      /* what about parallel ??? */
      if( fabs(a)<0.99 ) break;
    }
    if( j==N ) return; /* colinear */
    for(i=0;i<3;i++) xyz[1][i]=s[j][i]-a*xyz[0][i];
    for(a=0.0,i=0;i<3;i++) a+=xyz[1][i]*xyz[1][i];
    for(a=1/sqrt(a),i=0;i<3;i++) xyz[1][i]*=a;
    /* --- z axis perpentikular to s0 and s1 */
    for(i=0;i<3;i++) xyz[2][i]=xyz[0][(i+1)%3]*xyz[1][(i+2)%3]
                              -xyz[0][(i+2)%3]*xyz[1][(i+1)%3];
    /* --- nun alle spins in neue Koordinaten drehen */
    for(j=0;j<N;j++){
      for(i=0;i<3;i++)
      ss[i]=xyz[i][0]*s[j][0]
           +xyz[i][1]*s[j][1]
      	   +xyz[i][2]*s[j][2];
      for(i=0;i<3;i++) s[j][i]=ss[i];
      kart2kug(s[j],&r,wt+j,wp+j);    
    }
}

void norm_angles2(double hx){	/* wt ca. 0..1, wp 0..2 [Pi] */
    double ss[3],r,a,b,na,nb; int i;
    if( hx ) return;
    ss[0]=s[0][0];
    ss[1]=s[0][1];
    ss[2]=s[0][2];
    for(i=0;i<N;i++){
      r=s[i][0]*s[i][0]+s[i][1]*s[i][1]+s[i][2]*s[i][2]-1;
      if(fabs(r)>1e-10)fprintf(stderr,"error: S%d^2-1=%e\n",i,r); }
    a=ss[0];			/* Drehung um z -> x=1 y=0 */
    b=ss[1];
    r=sqrt(a*a+b*b);
    if(r>1e-20){
      r=1/r;
      na=r*( a*ss[0]+b*ss[1]);   /* rueckdrehung */
      nb=r*(-b*ss[0]+a*ss[1]);
      ss[0]=na;
      ss[1]=nb;
      for(i=0;i<N;i++){    
        na=r*( a*s[i][0]+b*s[i][1]);   /* rueckdrehung */
        nb=r*(-b*s[i][0]+a*s[i][1]);
        s[i][0]=na;
        s[i][1]=nb; } }
    a=ss[0];			/* Drehung um y -> z=0 x=1 */
    b=ss[2];
    r=sqrt(a*a+b*b);
    if(r>1e-20){
      r=1/r;
      na=r*( a*ss[0]+b*ss[2]);   /* rueckdrehung */
      nb=r*(-b*ss[0]+a*ss[2]);
      ss[0]=na;
      ss[2]=nb;
      for(i=0;i<N;i++){    
        na=r*( a*s[i][0]+b*s[i][2]);   /* rueckdrehung */
        nb=r*(-b*s[i][0]+a*s[i][2]);
        s[i][0]=na;
        s[i][2]=nb;  } }
    for(i=0;i<N;i++)kart2kug(s[i],&r,wt+i,wp+i);    
      }

/* ================================================================= */
double saved_s[N][3],saved_sum_sz;	/* spins and sum_sz */
int    saved_wt[N],saved_wp[N];		/* spin-vectors 0..MAXW=0..2*pi */
int save_s(){ int i,j;
  saved_sum_sz=sum_sz;
  for(i=0;i<N;i++){
     for(j=0;j<3;j++) saved_s[i][j]=s[i][j];
     saved_wt[i]=wt[i];
     saved_wp[i]=wp[i]; }
  return 0;}
          
int load_s(){ int i,j;
  sum_sz=saved_sum_sz;
  for(i=0;i<N;i++){
     for(j=0;j<3;j++) s[i][j]=saved_s[i][j];
     wt[i]=saved_wt[i];
     wp[i]=saved_wp[i]; }
  return 0;}          
/* ================================================================= */

Static_Inline
double myexp(double x){ /* exponent exp(x), x=-E/kT, x<MaxExp besser x<0 */
  if (x < -MaxExp) return 0.0;
  return exp(x);}

/* hamiltonien ({0,j1,j2,j3}) */
double ham(double jj[NJ+1][3], double hx){
  double r=0.0,*rj,*s1,*s2; int i;      /* evl. non-0-Tabellen nutzen */
  for(i=0;i<ww_n;i++){
    s1= s[ ww_i[i] ]; /* site 1 */
    s2= s[ ww_j[i] ]; /* site 2 */
    rj=jj[ ww_r[i] ]; /* JJ-index */
    r += rj[0]*s1[0]*s2[0]
       + rj[1]*s1[1]*s2[1]
       + rj[2]*s1[2]*s2[2];
  }
  /* apply an externel field +hx*Sx */
  if(hx!=0)  for(i=0;i<N;i++){  r += 2*s[i][0]*hx; }
  return 0.5*r;}  /* why 0.5 ??? */

double emax(double jj[NJ+1][3]){     /* hamiltonien ({0,j1,j2,j3}) */
  double r=0.0,*rj; int i;
  for(i=0;i<ww_n;i++){
    rj=jj[ ww_r[i] ];
    r+=sqrt(rj[0]*rj[0]+rj[1]*rj[1]+rj[2]*rj[2]);} 
    /* |J| because prd()=-1..+1 */
    /* externel field */
    r+=2*N*fabs(hx);
  return 0.5*r;}

/* eff. UmgebungsSpin am Platz i berechnen und in *su speichern */
Static_Inline
double *umgebung(int i,double su[3]){
  int j;double *sp,*rj;
  su[0]=su[1]=su[2]=0.0;
  for(j=ww_a[i];j<ww_b[i];j++) {
    sp= s[ ww_j[j] ];
    rj=jj[ ww_r[j] ];
    su[0]+=rj[0]*sp[0];
    su[1]+=rj[1]*sp[1];
    su[2]+=rj[2]*sp[2]; }
  return(su);}

long cluster_size=0l;
/*
 - dieser Algorithmus funktioniert nicht bei anisotropie! 
   keine detailierte balance!!
 - Cluster flippen, REKURSIV aufgerufen,
   besser Rekursion entflechten -> Sedwick: Algorithmen etc.
 - ToDo: check Gitter mit Basis???
 */
Static_Inline   		/* bei rekursiv moeglich?? */
void flippcluster(int i){	/* sp=spiegelvektor */
  static int flag[N],  	/* flag=1  flipp spin, initialize with 0! */
    list[N], /* liste der zum cluster gehoerenden spins, oder pointerliste */
    li,le;   /* list_index, list_end */
  double sn[N],sp[3]; 	/* store (s[i],n), sp=vec_n */
  unsigned h2,h3;	/* Dez98 vorher double! */
  int j;
  double ri,*rj,rr,*sj,*si,pp;
  si=s[i]; flag[i]=1;
#if 1			
  h2 = rand()&(MAXW-1);	/* original cluster, n=random */
  h3 = rand()&(MAXW-1);	/* korrekte Verteilung! cos^-1(-1..1) */
  setspin(sp,h2,h3);	/* spiegelvektor */
  ri=sp[0]*si[0]+sp[1]*si[1]+sp[2]*si[2];                   /* (si,n) */
#else			/* modified cluster, n=s[random] */
  sp[0]=si[0]; sp[1]=si[1]; sp[2]=si[2]; ri=1.0;            /* (si,n) */
#endif
  sn[i]=ri;list[0]=i;li=0;le=1;		       /* speichern fuer flip */
  do {
    i=list[li];ri=sn[i];
    for(j=ww_a[i];j<ww_b[i];j++)                  /* nachbarn untersuchen */
    if(!flag[ww_j[j]]) {                  	   /* noch nicht geflippt */
      sj= s[ ww_j[j] ];                                    /* nachbarspin */
      rj=jj[ ww_r[j] ];                                /* bound, kopplung */
      rr=sn[ ww_j[j] ]=sp[0]*sj[0]+sp[1]*sj[1]+sp[2]*sj[2];     /* (sj,n) */
      rr=rj[0]*ri*rr;				       /* J*(si,n)*(sj,n) */  
      /* FM if parallel, flipp with Pad=pp */
      /* 2*rr=dEij=Eij(nachflip)-Eij(vorflip)=2*jj*(si,n)*(sj,n)
                  =   jj*(Si,Sj')        -jj*(Si,Sj)
                  =   jj*(Si,Sj-2(Sj,n)n)-jj*(Si,Sj)
                  =-2*jj*(Si,n)*(Sj,n)             => -2*(JJ x Si,n)*(Sj,n) */
      if(rr<0) { /* FM: wenn parallel mit probpability Padd zufuegen; kT=0? */
         pp=1-myexp(2*beta*rr);
         if(pp>=_randreal()){
           list[le]=ww_j[j];flag[ ww_j[j] ]=1;
           if(le<N)le++;
         }
      }  
    }
    li++;  
  } while(li<le);                          /* li<le<=N*/  
  for(li=0;li<le;li++){	/* bei cluster>N/2 evl. restgitter flippen! */ 
    i =list[li];
    ri=2*sn[i];           /* flippen p'=p-2n(n,p)/(n,n) */
    s[i][0]-=ri*sp[0];
    s[i][1]-=ri*sp[1];
    s[i][2]-=ri*sp[2];
    sum_sz -=ri*sp[2];
    for(j=ww_a[i];j<ww_b[i];j++)                  /* nachbarn untersuchen */
    if(!flag[ww_j[j]]) {                  	   /* noch nicht geflippt */
      rj=jj[ ww_r[j] ];                                /* bound, kopplung */
      gerg-=rj[0]*ri*sn[ ww_j[j] ];		  /* -2*J*(d_si,n)*(sj,n) */  
    }
  }
  for(li=0;li<le;li++){	/* flags loeschen */ 
    i =list[li]; flag[i]=0;
  }
  cluster_size=le;  
  return;}


Static_Inline double dham(int i,double sn[3]){
  double su[3];	/* neuer Spin Sn, UmgebungsSpin Su von i */
  /* eff. UmgebungsSpin berechnen! evl. in Array speichern? */
  umgebung(i,su);
  return( (sn[0]-s[i][0])*su[0]+ 	/* skalarprd. (Sn-Si)*Sui */
  	  (sn[1]-s[i][1])*su[1]+	/* oder hier sn optimieren! */
  	  (sn[2]-s[i][2])*su[2]+
  	  (sn[0]-s[i][0])*hx );		/* field correct??? */
}

void outspin(FILE *f1,double *s){
  double r; int t,p;
  kart2kug(s,&r,&t,&p);
  fprintf(f1," %8.6f %4.2f %4.2f",r,t/(.5*MAXW),p/(.5*MAXW));
}

/* summation of sublattice spins and total spin */
void sublattice( FILE *f1 ){
  double ssub[N][3],ss[3]; int i,j,nsub[N];
  for(i=0;i<ini_nsl;i++){ nsub[i]=0; for(j=0;j<3;j++)ssub[i][j]=0; }
  ss[0]=ss[1]=ss[2]=0;
  for(j=0;j<N;j++){
     nsub[ini_sl[j]]++;
     for(i=0;i<3;i++){
       ssub[ini_sl[j]][i]+=s[j][i];
       ss[i]+=s[j][i];
     }
   }
   if(ini_nsl>1){
     fprintf(f1," S_SL=");
     for(i=0;i<ini_nsl;i++){
       for(j=0;j<3;j++)ssub[i][j]/=nsub[i]; /* normalize */
       outspin(f1,ssub[i]); 
     }
   }
   fprintf(f1," S=");
   for(j=0;j<3;j++)ss[j]/=N; /* normalize */
   outspin(f1,ss);
}

#define SCL 500	/* scale */
/* --- old: load "state.out";plot "lattice.out"   --- */
/* --- new: mcarlo ... -v2 | gawk -f raw2fig.awk > a.fig ; xfig a.fig --- */
int outstate( FILE *f1 ){
 double l=.25;
 int i,j,k; double pos[3],po2[3],xh,xl,yh,yl,zh,zl;
 xh=yh=zh=-99;xl=yl=zl=99;
 for(i=0;i<N;i++){
    site_pos(i,pos);
    if(pos[0]>xh)xh=pos[0]; /* maxima */
    if(pos[1]>yh)yh=pos[1];
    if(pos[2]>zh)zh=pos[2];
    if(pos[0]<xl)xl=pos[0]; /* minima */
    if(pos[1]<yl)yl=pos[1];
    if(pos[2]<zl)zl=pos[2];
  }
  fprintf(f1,"# raw output of mcarlo describing spin state\n");
  fprintf(f1,"box %4.2f %4.2f %4.2f  %4.2f %4.2f %4.2f\n",
    xl, yl, zl, xh, yh, zh);
  for(i=0;i<N-1;i++)
  for(j=i+1;j<N;j++){  /* dont draw bonds twice */
    k=ww_get(i,j);
    if(k){
      site_pos(i,pos);
      site_pos(j,po2);
#if 0
      /* dont draw PBC-bounds, that fails on exotic structures */
      /* better do it in the awk-script? or use (ini_wa[k]&1)==0 */
      if(fabs(pos[0]-po2[0])<3 && fabs(pos[1]-po2[1])<3)
#endif
      fprintf(f1,"line %5.2f %5.2f %5.2f  %5.2f %5.2f %5.2f  % 4.2f # ri rj sisj\n",
        pos[0],pos[1],pos[2],po2[0],po2[1],po2[2],prd(i,j) );
    }
  }
 for(i=0;i<N;i++){
    site_pos(i,pos);
    fprintf(f1,"spin %5.2f %5.2f %5.2f  %5.2f %5.2f %5.2f %3d"
               "  %7.4f %7.4f # r s n  t p\n",
     pos[0],  pos[1],  pos[2],
    s[i][0], s[i][1], s[i][2], i, wt[i]/(.5*MAXW), wp[i]/(.5*MAXW));
  }
  return 0;
}
  
/*
 * spins local transformieren, damit s_0=(0,0,1) !? korrelationen, magnetis.
 *  changed to title= data (more size, but easier to read and awk possible)
 */
void outall( FILE *f1 ){
    int i,j,i1;	/* Teil-Energien */
    static double jj1[NJ+1][3];
    double rr;
    /* auf (t,p)_0=(0,0) transformieren, t=0..1 p=0..2 (p>0)!
     * if ((plane parallel to hx) or hx==0) and isotropic jj ???
     */
    for(i1=0;i1<NJ+1;i1++)
      if (jj1[i1][0]!=jj1[i1][1]
       || jj1[i1][0]!=jj1[i1][2]) break;
    if ( i1==NJ+1 ) norm_angles(hx);
    /* dipolmoment in hx Richtung ??? */
    /* quadrupelmoment in x+y Richtung drehen ??? */
    fprintf(f1," hx= %4g J= %2g %4g %5g e= %14.9f",
      hx,jj[1][2],jj[2][2],jj[3][2],ham(jj,hx)/N);
    for(i=1;i<4;i++){
      for(i1=0;i1<NJ+1;i1++){ jj1[i1][0]=jj1[i1][1]=jj1[i1][2]=0; }
      jj1[i][0]=jj1[i][1]=jj1[i][2]=1;
      fprintf(f1," e%d= %9.6f",i,ham(jj1,0)/N);
    }
    sublattice( f1 );

    if (vvv & 2) outstate(stdout);

    for(i=0;i<1;i++)for(j=0;j<N;j++)
    fprintf(f1," S%dS%d= %9.6f",i,j,prd(i,j));

    for(rr=0,i=0;i<N;i++) rr+=prd(i,i);
    fprintf(f1,    " (Si)^2= %9.6f",rr/(N));

     for(rr=0,i=0;i<N;i++)for(j=0;j<N;j++) rr+=prd(i,j)*prd(i,j);
    fprintf(stderr," (SiSj)^2= %9.6f",rr/(N*1.*N));
    fprintf(f1    ," (SiSj)^2= %9.6f",rr/(N*1.*N));

    for(rr=0,i=0;i<N;i++)for(j=0;j<N;j++) rr+=fabs(prd(i,j));
    fprintf(stderr," |(SiSj)|= %9.6f\n",rr/(N*1.*N));
    fprintf(f1    ," |(SiSj)|= %9.6f\n",rr/(N*1.*N));

    /* colinearer Zustand? (Jan01 JS) = alle skalare \pm 1 */
    rr=1.0;
    for(i=1  ;i+1<N;i++) if( fabs(prd(0,i))<rr ) rr=fabs(prd(0,i));
    fprintf(f1," min|V0.Vi|= %.1e %s",rr,((rr>1-1e-5)?"colinear":"-       "));

    /* planarer Zustand? (Mar99 JS) = alle spatprodukte 0 */
    rr=0.0;
    for(i=1  ;i+1<N;i++)
    for(j=i+1;j  <N;j++) if( fabs(spat(0,i,j))>rr ) rr=fabs(spat(0,i,j));
    fprintf(f1," max|(V0xVi)Vj)|= %.1e %s",rr,((rr<4e-5)?"planar":"3d    "));

    fprintf(f1,"\n");fflush(f1);
}
/* ------- class mean-values class thermodynamik ------------- */
/* uses: gerg,gkt,N */
#define HistoN 256		/* should be 2^n */
double z,z1,z2,z3,z4,z5,z6,z7,z8,z9,zc,ze,zee,zz,zzz,zz4;
long hzz[HistoN],hzc[HistoN];

/* --- zur Abkuerzung Makro definieren, nur fuer mean_out --- */
#define FN1(A,B) fprintf(ff," "A,(B)) 
void mean_out(FILE *ff){ 
#ifdef CONFIG_HISTO
   FILE *f1;
#endif   
   FN1(  "<E/N> %8.4g",(z)?ze /( z*N):0); /* energy */
   FN1( "<cs>/N %8.6f",(z)?(zc)/(N*z):0); /* cluster_size of flippcluster() */
   FN1(    "c/N %8.6f",(z*gkt)?(z*zee-ze*ze)/(N*z*z*gkt*gkt):0); /* spec. heat */
   FN1(  "<Z>/N %8.5f",( z)? zz/(N*z):0); 
   FN1("<Z^2>/N %8.5f",( z)?zzz/(N*z):0); /* chi=(<Z^2>-<Z>^2)/kT */ 
   FN1("<Z^4>/N %8.5f",( z)?zz4/(N*z):0); /* Binder parameter U_L */
					    /* U_L=1-<z^4>/3<z^2>^2 */
   FN1(   "S0S1 %9.5f",( z)?(z1/z):0); 
   FN1(   "S2S3 %9.5f",( z)?(z2/z):0);
#if defined(CONFIG_HISTO) /* Histo */
   {
     f1 = fopen("histo.out" ,"w"); if(!f1){ fprintf(stderr,"OpenError1\n");exit(1);}
     fprintf(f1,"# kT=%g\n",gkt);
     for(i=0;i<HistoN;i++){ 
       fprintf(f1,"%6.4f %e %e\n",i*1./HistoN,hzz[i]/z,hzc[i]/z); }
     fclose(f1);
   }
#endif
  }
#undef FN1

void mean_ini(){ int i; 
  z1=z2=z3=z4=z5=z6=z7=z8=z9=zz=zzz=zz4=zc=z=ze=zee=0.0; /* ini_erg */
  for(i=0;i<HistoN;i++) hzz[i]=hzc[i]=0l;
  sum_sz=0.0; for(i=0;i<N;i++) sum_sz+=s[i][2];
 }

void mean_calc(){ 
   zz += sum_sz;		/* nur bei flipp geaendert */
   zzz+= sum_sz*sum_sz; 
   zz4+= sum_sz*sum_sz*sum_sz*sum_sz; 
   ze += gerg;          	/* => system.get_erg exp(a/b) */
   zee+= gerg*gerg;		/*  */
   z1 += prd(0,1);		/* schnellerer E-Ersatz */
   z2 += prd(2,3);		/* schnellerer C-Ersatz ??? */
   z  +=        1;
   zc +=cluster_size;
#if defined(CONFIG_HISTO)
   hzz[ (int)( HistoN*fabs(sum_sz) /N)&(HistoN-1) ]++;
   hzc[ (int)((HistoN*cluster_size)/N)&(HistoN-1) ]++;
#endif
}
   

/* ------------------------------------------------------------------- */

void out2(FILE *ff){
  fprintf(ff," kT= %8.5g",gkt); mean_out(ff);
  fprintf(ff,"\n"); }

void outE(FILE *ff){
   /* fprintf(ff,"\r"); */
   fprintf(ff," e0= % .6e",gerg);			
   fprintf(ff," e0/w= % .6e",2*gerg/ww_n);			
   if(anz>999999) fprintf(ff," n= %6luk",anz/1000);             
   else           fprintf(ff," n= %7lu",anz);
   if(gkt  !=0.0) fprintf(ff," T= %5.3f",gkt);  	       	
   if(z*zc !=0.0) fprintf(ff," cs= %7.5f",zc/(N*z)); /* cluster_size */
   if(z*ze !=0.0) fprintf(ff," e= %7.4f",ze/(N*z));  /* energy <H> */        	
   if(z*gkt*ze*zee!=0.0)
   fprintf(ff," c= %7.4f",(z*zee-ze*ze)/(N*z*z*gkt*gkt)); /* spec. heat */

   if(ifNaN(gerg))fprintf(ff," error 1\n");  /* NaN ? */
   fflush(ff);}

/* ersten Spin nicht festsetzen! sonst Konvergenzprobleme!??? */
void monte2(){  		/* Monte Carlo Methode */
    double e0, d1, sn[3];
    int i1,h2,h3; 

    beta = 1/gkt;  e0 = e_border;

    mean_ini();
    gerg = ham(jj,hx);

    for(anz=1;anz<steps;anz++) {
	
	i1=rand()%N;  /* rand() and %() needs lot of time */	
	/* ====== spin zufaellig varrieren ======== */
	h2 = rand()&(MAXW-1);
	h3 = rand()&(MAXW-1);
	setspin(sn,h2,h3);

	d1=dham(i1,sn);  /* hneu-hold(=gerg) */

/*	if ((d1<=0.0) || (myexp(-d1*beta) >= _randreal()) ) { */
	if ((d1<=0.0) || (-d1*beta >= logx[rand()&(MAXW-1)] )) { /* 0..-10 */
	  /* besserer Zustand */
  	  wt[i1]   = h2;
	  wp[i1]   = h3;
	  s[i1][0] = sn[0];
	  s[i1][1] = sn[1];sum_sz-=s[i1][2];
	  s[i1][2] = sn[2];sum_sz+=s[i1][2];
	  gerg    += d1;  }
	
	if(anz>start) mean_calc();	/* Erwartungswerte */ 
    }
    if(vvv){outE(stderr);fprintf(stderr,"\n");}
}

/*         
   importent MC  
             -> a) 1 Spin aendern wenn exp(-beta*(Eneu-Ealt))<random(0..1)
                b) ZE += E 
                   Z  += 1
                   c   = ZE / Z
                c) goto a) 

   simple MC -> a) neuer zufaelliger Zustand (1 Spin)
                b) ZE += E * exp ( -beta*E )
                   Z  +=     exp ( -beta*E )
                   c   =  ZE / Z
                c) goto a) 
*/
/* Cluster-Flip-Methode, geht nur bei isotropen Hamiltonien */
/* what about hfield ??? */
void monte3(){  		/* Monte Carlo Methode */
    double e0;    int i1;

    beta = 1/gkt;  e0 = e_border;
    mean_ini();    gerg = ham(jj,hx);

    for(anz=1l;anz<start;anz++) {			/* warm up */
      	i1=rand()%N;	flippcluster(i1);  }

    for(anz=1l;anz<steps;anz++) {			/* measure */
      	i1=rand()%N;	flippcluster(i1);
	/* if(!(255&(int)anz)) gerg=ham(jj,hx); */ /* zeit sparen !? */
      	mean_calc();  
      	/* alle 2^x schritte genauigkeit pruefen!? sum_sz, erg, norm(s) */
    }
      	
    if(vvv){outE(stderr);fprintf(stderr,"\n");}
}


/* "Abschrecken" -> schnelle Abkuehlung, minimale Energie */
/* hx-field correct ??? */
void method1(){
    double r,gerg1,d1,sn[3]; int i;

    gerg1 = gerg = ham(jj,hx);

    for(anz=1;;anz++){			/* alle Spins */

      for (i=0; i<N; i++) {
        /* oder zum Nachbarn mit groesster lok. Energie huepfen */

      	/* ====== spin energetisch optimieren ======== */
        umgebung(i,sn);
        r = sn[0]*sn[0] + sn[1]*sn[1] + sn[2]*sn[2];  /* Normiere */
        sn[0]+=hx;     /* field right implemented ??? */
        r = sn[0]*sn[0] + sn[1]*sn[1] + sn[2]*sn[2];  /* Normiere */
        if(r>1e-12){       /* norm==0 --> optimaler Spin beliebig */
          d1=-1/sqrt(r);   /* optimaler Spin = -umgebung/norm */
          sn[0]*=d1; sn[1]*=d1; sn[2]*=d1;

   	  d1=dham(i,sn);        /* hneu-hold(=gerg) */
	  if (d1<=0.0)  {	/* besserer Zustand */
	    s[i][0] = sn[0];
	    s[i][1] = sn[1];sum_sz-=s[i][2];
	    s[i][2] = sn[2];sum_sz+=s[i][2];
	    gerg   += d1;
          }
        }
      }
      if(gkt>0  && gerg1-gerg<1e-6) break;  /* enough precission */
      if(gkt==0 && gerg1-gerg<1e-9) break;  /* enough precission? */
      if (gerg<gerg1) gerg1=gerg; else break;
    }
    if(vvv){outE(stderr);fprintf(stderr,"\n");}

    for (i=0; i<N; i++) kart2kug(s[i],NULL,&wt[i],&wp[i]);
}

int ini_spins(int i){  /* Startzustand i: 0=random */
#if N>100000
    double pos[3];
#endif    
    sum_sz=0.0;
    for (i=0; i<N; i++){
#if N>100000		/* GS of very large unfrustrated cubic lattices */
      site_pos(i,pos);
      wt[i]=(((int)(pos[0]+pos[1]+pos[2]+.5)%2)?0:1);wp[i]=0;
      s[i][0]=s[i][1]=0.0;
      s[i][2]=(((int)(pos[0]+pos[1]+pos[2]+.5)%2)?1.0:-1.0);
#else     
      setrandomspin(i);
#endif
      sum_sz+=s[i][2];
    } return 0; }    

int ini_bonds(){  /* ww -> ww_n,ww_i,ww_r  non-0 Elemente fuer schnelleres Rechnen */
  int i,j; double x;
  ww_n=0;
  for(i=0;i<N;i++){  /* N*Neighbours better than N*N time consuming */
    ww_a[i]=ww_n;
    for(j=0;j<ww_num_neighbours(i);j++){
      ww_i[ww_n]=i;
      ww_neighbour(i,j,ww_j+ww_n,ww_r+ww_n);/* set ww_j, ww_r */
#if 0
      fprintf(stderr,"i=%u %u--%u typ=%u\n",ww_n,ww_i[ww_n],ww_j[ww_n],ww_r[ww_n]);
#endif
      ww_n++; 
      if(ww_n>8*N){ fprintf(stderr,"L%d ww_n>8*N\n",__LINE__);exit(1); }
      }
    ww_b[i]=ww_n; }
  for(i=0;i<MAXW;i++){
    x=2*i*M_PI/MAXW;			/* phi */
    sinx[i]=sin(x);    cosx[i]=cos(x);
    x=acos(i*2.0/MAXW-1.0);		/* wegen Gleichverteilung theta */
    siny[i]=sin(x);    cosy[i]=cos(x);
    x=i/(double)MAXW;
    logx[i]=(x>0.0)?log(x):-exp(MaxExp); }
  return 0; }
  
int help(int err){
  printf("mcarlo  version " version_string "  (w) Joerg Schulenburg\n");
  printf(" use:\n"
    "  mcarlo [options] -j1=<val> [-j2=<val> ...]\n"
#if 0
    "      kill -SIGUSR1 <pid>    \t-> do_cmd_i;i=0;\n"
    "      kill -SIGUSR2 <pid>    \t-> i++;\n"
    "      cmd_0 = print out status of program\n"
#endif
    " options:\n"
    "   -h<field>       - set h field (x-direction)\n"
    "   -j<n>=<jj>      - set exchange j<n> to jj (n=1..%d)\n"
    "   -j<n>=<jx>,<jy>,<jz> - set exchange j<n>=(jx,jy,jz)\n"
    "   -r<srand>       - initialize randomgenerator\n");
  printf(
    "   -s<steps>       - numSteps!=1M\n"
    "   -t<temp>        - set temperature (value or range Tmin-Tmax)\n"
    "   -v<n>           - be verbose (1=heads,2=state,4=dbg,8=...)\n"
    "   -q              - be quite (-v0)\n"
    " note: Field hx is not proper implemented, use with care!\n",NJ
  );
  if ( err ) exit( err );
  return 0;
}

int main(int argc,char *argv[]){
  int i,j,ini;
  double kT1,kT2,r=0.0;
  char *s1,*s2;

  kT1=kT2=gkt=0.0;		/* Temperatur; kt -->0 Grundzustand */

  for(i=0;i<NJ+1;i++) jj[i][2]=jj[i][1]=jj[i][0]=0;
  if(argc<2) help(1);
  j=0; ini=0;
  for( j=1; j<argc; j++ ){ 
    s1=argv[j]; if(s1[0]!='-') break;
    if(s1[1]=='q'){ vvv=0; } else
    if(s1[1]=='r'){ ini=atoi(s1+2); } else
    if(s1[1]=='v'){ vvv=atoi(s1+2); } else
    if(s1[1]=='s'){ steps=atol(s1+2); } else
    if(s1[1]=='h'){ hx=atof(s1+2); } else
    if(s1[1]=='j'){
      i = atoi( s1+2 ); if( i<1 || i>NJ ) F1("only j1 ... j%d allowed",NJ);
      s2=strchr( s1+2, '=' );
      if (!s2) F1("expect -j%d= but did not found =",i);
      jj[i][2]=jj[i][1]=jj[i][0]=atof(s2+1);
      if (s2) s2=strchr(s2+1,','); if (s2) jj[i][2]=jj[i][1]=atof(s2+1);
      if (s2) s2=strchr(s2+1,','); if (s2) jj[i][2]=         atof(s2+1);
      if( vvv & 4 )
       fprintf( stderr, " J%d=(%g,%g,%g)\n",i,jj[i][0],jj[i][1],jj[i][2] );
    } else
    if(s1[1]=='t'){
      kT1=kT2=gkt=atof(s1+2);
      s2=strchr( s1+2, '-' ); if ( s2 ) { kT2 = atof(s2+1); }
      if( vvv & 4 ) fprintf( stderr, " T=%g-%g\n",kT1,kT2 );
    } else
    help(1);
  }
  f1 = fopen("out" ,"a"); if(!f1){ fprintf(stderr,"OpenError1\n");exit(1);}
  setvbuf(  stdout, (char *)NULL, _IONBF, 0);	/* nicht puffern!	*/
  if(f1)setvbuf(f1, (char *)NULL, _IONBF, 0);	/* nicht puffern!	*/

  /* if(N==12) jj[2]*=2;   halbe Bindungszahl ausgleichen (2*2) */
  if( vvv & 1 )
  fprintf(f1,"# mcarlo Jz1=%g Jz2=%g Jz3=%g Ja3=%g kT=%g..%g steps=%.0e..%.0e N=%lu hx=%g XY=%d\n",
    jj[1][2],jj[2][2],jj[3][2],jj[3][0]-jj[3][2],kT1,kT2,
    (double)start,(double)steps,(unsigned long)N,hx,XY);
  ww_ini();			/* Modell laden/initialisieren */
  ini_bonds();			/* J_ij besser speichern */
  _randomize(ini);
  ini_spins(0);			/* initialize state, 0=rand or 1=GS? */
  e_border=emax(jj);		/* upper border for energy (eq lower border) */
  ini_handler(vvv);		/* allow interrupts */
  if(kT2==0.0){  
    method1(); outall(f1);	/* T=0, optimize */
    /* monte3(); outall(f1); */ /* only for tests !!! */
  } else {			/* theta=0.5*M_Pi,phi=monte_carlo (XY) */
   method1();save_s();
#if 0
   /* to get an overview as fast as possible the grid will be
    * 0.5   0.25 0.75   0.125 0.375 0.625 0.875  ....
    * disadvantage: no heat up or cool down possible
    *   only 2^steps (64) possible    */
   for(i=1;i<64;i*=2) { 	/* kT as grid becoming more fine */
    for(j=0;j<i;j++) {
      gkt=kT1+(kT2-kT1)*(j+.5)/i;  	/* start=random or T=0 GS */
#else
   for(i=0;i<64;i++) { { 	/* simple variant, 64 steps */
      gkt=kT1+(kT2-kT1)*i/64.;  /* start=random or T=0 GS */
#endif
      load_s();
      if(hx==0) monte3(); 	/* T>0, cluster flip works for h=0 */
      else      monte2(); 	/* T>0, old algo (slow but ok for h>0?) */
      out2(f1);
    }
    if(kT2==kT1)break; /* single temperatur */
   }
  }
  if (f1) fclose(f1);
  return 0;
}

